# Created by: Ausar686
# https://github.com/Ausar686

import os
from collections import defaultdict

import numpy as np

from ..storages import Actions, Diary, MediaStorage, UserData 


class Profile: 
    _user_data_filename = "user.json"
    _diary_filename = "diary.csv"
    _actions_filename = "actions.csv"
    _media_data_filename = "media.csv"
    
    def __init__(
        self,
        load_dir: str=None,
        *,
        user_data: UserData=None,
        diary: Diary=None,
        actions: Actions=None,
        media: MediaStorage=None,
        default_dir: str=None):
        if load_dir is not None:
            self.load(load_dir)
            return
        if isinstance(user_data, UserData):
            self.user_data = user_data
        else:
            self.user_data = UserData()
        if isinstance(diary, Diary):
            self.diary = diary
        else:
            self.diary = Diary()
        if isinstance(actions, Actions):
            self.actions = actions
        else:
            self.actions = Actions()
        if isinstance(media, MediaStorage):
            self.media = media
        else:
            self.media = MediaStorage()
        if default_dir is None:
            self.default_dir = load_dir
        else:
            self.default_dir = default_dir
        self.up_to_date = True
        return
    
    def get_emotions_data(self, value: str=None) -> dict:
        df = self.diary.get(by="emotion", value=value)
        if df.empty:
            return {}
        values = df["reason"].value_counts()
        emotions_data = {key: values[key] for key in values}
        return emotions_data
    
    def get_media_data(self, media: str=None) -> dict:
        df = self.actions.get(by="media", value=media)
        if df.empty:
            return {}
        values = df[["title", "author"]].value_counts()
        media_data = dict(values)
        return media_data
    
    def add_emotion(self, emotion: dict) -> None:
        self.diary.append(emotion)
        self.up_to_date = False
        return
    
    def add_action(self, action: dict) -> None:
        self.actions.append(action)
        self.up_to_date = False
        return
    
    @staticmethod
    def _score_function(array: np.array) -> np.array:
        return np.exp(0.25*(array-array.shape[0]))
    
    def _update_actions_data(self) -> None:
        self.actions.data["score"] = self._score_function(self.actions.data.index)
        self.actions.data["_topic"] = self.actions.data.topic.str.split("|").fillna(self.actions.data.topic)
        self.actions.data = self.actions.data.explode("_topic") # Check this
        self.user_data._fav_actions_topics = defaultdict(float, [(topic, df.score.sum()) 
                                                        for topic, df in self.actions.data.groupby(by="_topic")])
        return
        
    def _update_diary_data(self) -> None:
        self.diary.data["score"] = self._score_function(self.diary.data.index)
        self.diary.data["_emotion"] = self.diary.data.emotion.str.split("|").fillna(self.diary.data.emotion)
        self.diary.data = self.diary.data.explode("_emotion") # Check this
        self.user_data._fav_diary_topics = defaultdict(float, [(topic, df.score.sum()) 
                                                     for topic, df in self.diary.data.groupby(by="_emotion")])
        return
        
    def _update_fav_topics(self) -> None:
        topics = set(self.user_data._fav_diary_topics.keys()) | set(self.user_data._fav_actions_topics.keys())
        self.user_data._fav_topics = \
            defaultdict(float, [(topic, self.user_data._fav_diary_topics[topic] + self.user_data._fav_actions_topics[topic]) 
                                for topic in topics])
        return
    
    def _update_media_data(self) -> None:
        self.media.data["_topic"] = self.media.data.topic.str.split("|").fillna(self.media.data.topic)
        self.media.data = self.media.data.explode("_topic") # Check this
        self.media.data["score"] = self.media.data["_topic"].map(self.user_data._fav_topics)
        summarized_df = self.media.data.groupby(level=0)["score"].sum()
        self.media.data.drop(["_topic", "score"], axis=1, inplace=True)
        self.media.data.drop_duplicates(inplace=True)
        self.media.data["score"] = summarized_df
        return
    
    def update(self) -> None:
        if self.up_to_date:
            return
        # Update actions' scores
        self._update_actions_data()
        # Update diary scores
        self._update_diary_data()
        # Update favourite topics in user_data._fav_topics defaultdict.
        self._update_fav_topics()
        # Update media relevance scores
        self._update_media_data()
        # Set the flag so that not to update data before actual action or diary note.
        self.up_to_date = True
        return
    
    def load(self, load_dir: str=None) -> None:
        if load_dir is None:
            raise ValueError("Can't load from None directory.")
        # Set paths to files with data
        user_data_path = os.path.join(load_dir, self._user_data_filename)
        diary_path = os.path.join(load_dir, self._diary_filename)
        actions_path = os.path.join(load_dir, self._actions_filename)
        media_data_path = os.path.join(load_dir, self._media_data_filename)
        # Create UserData from JSON
        self.user_data = UserData(user_data_path)
        # Create Diary from path
        self.diary = Diary(diary_path)
        # Create Actions from path
        self.actions = Actions(actions_path)
        # Create MediaStorage from path
        self.media = MediaStorage(media_data_path)
        # Set default directory for further 'to_disk' and 'load' usage
        self.default_dir = load_dir
        # Set 'up_to_date' attribute for updating
        self.up_to_date = False
        # Update the data in profile
        self.update()
        return
    
    def to_disk(self, output_dir: str=None) -> None:
        if output_dir is None:
            output_dir = self.default_dir
        if not os.path.exists(output_dir):
            raise ValueError("Output directory does not exist.")
        # Set paths to files with data
        user_data_path = os.path.join(output_dir, self._user_data_filename)
        diary_path = os.path.join(output_dir, self._diary_filename)
        actions_path = os.path.join(output_dir, self._actions_filename)
        media_data_path = os.path.join(output_dir, self._media_data_filename)
        # Write data to disk
        self.user_data.to_disk(user_data_path)
        self.diary.to_disk(diary_path)
        self.actions.to_disk(actions_path)
        self.media.to_disk(media_data_path)
        return


# Created by: Ausar686
# https://github.com/Ausar686

import json


class Poll:
    
    _bot = "[RBot]"
    
    def __init__(self, json_path: str) -> None:
        self.poll = None
        self.load(json_path)
        self.user_answers = {}
        self.joiner = '\n\t'
        self.instructions = f"""
            <INSTRUCTIONS>
            Write a conclusion about my current condition, using the information below.
            Your answer should be provided in Russian language.
            Use "ты" in order to address to me.
            <CONTEXT>
            """
        self._str = self.to_str()
        return
        
    def load(self, json_path: str) -> None:
        with open(json_path, "r") as json_file:
            self.poll = json.load(json_file)
        return
    
    def run(self) -> None:
        for elem in self.poll:
            question, variants, answers = self.unpack(elem)
            self.display(question, answers)
            user_answer = self.input()
            self.user_answers[question] = user_answer
            try:
                index = answers.index(user_answer)
            except ValueError:
                index = -1
            if index < 0:
            # Improve this in further updates.
                func_name = variants[-1]["call"]["name"]
                eval(f"self.{func_name}()")
                continue
            variant = variants[index]
            reply = variant.get("reply")
            if reply:
                print(f"{self._bot}: {reply}")
            call = variant.get("call")
            if call:
                func_name = call.get("name")
                kwargs = call.get("kwargs")
                if kwargs:
                    eval(f"self.{func_name}(**{kwargs})")
                else:
                    eval(f"self.{func_name}()")
                continue
        self.instructions += str(self.user_answers)
        return
                
    def unpack(self, elem: dict) -> tuple:
        question = elem["question"]
        variants = elem["variants"]
        answers = [variant["answer"] for variant in variants if variant["answer"]]
        return (question, variants, answers)
    
    def format(self, question: str, answers: list) -> str:
        return f"{question}{self.joiner}{self.joiner.join([f'{i+1}.{answer}' for i, answer in enumerate(answers)])}"
                
    def display(self, question: str, answers: list) -> None:
        print(f"{self._bot}: {self.format(question, answers)}")
        return
                
    def input(self) -> str:
        return input("[User]: ")
    
    def to_str(self) -> str:
        res = ''
        for elem in self.poll:
            question, variants, answers = self.unpack(elem)
            res += self.format(question, answers) + '\n'
        return res

    def analyse_input(self) -> None:
        print(f"{self._bot}: Я только учусь воспринимать человеческий текст в мини-тестах и опросах.")
        print(f"{self._bot}: Я записал твой ответ, но, пожалуйста, постарайся в будущем отвечать, используя предоставленные варианты.")
        return

    def ask_condition(self, condition: str) -> None:
        print(f"{self._bot}: Ты сейчас испытываешь состояние '{condition}'\n\tРасскажи, пожалуйста, поподробнее, чтобы я мог тебе помочь.")
        descr = self.input()
        self.user_answers[condition] = descr
        print(f"{self._bot}: Спасибо! Я немного подумаю, как тебе помочь и чуть позже напишу!")
        return
    
    def __repr__(self) -> str:
        return self._str
    
    def __str__(self) -> str:
        return str(self.poll)


# Created by: Ausar686
# https://github.com/Ausar686

from .anti_list_actor import AntiListActor
from .chat_bot import ChatBot
from .knowledge_base_searcher import KnowledgeBaseSearcher
from .qagpt import QAGPT
from .token_counter import TokenCounter
from .text_summarizer import TextSummarizer


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union, Any
from datetime import datetime
import json


class Message:
    """
    Class, that implements messages in chat.
    """
    
    _attributes = ["content", "role", "username", "chat_id", "date", "time", 
                   "is_reply", "reply", "is_forward", "forward", "media"]
    _bot_data = ['role', 'content']
    _front_keys = ["content", "role", "username", "date", "time"]
    
    def __init__(self, content: Union[str,dict,"Message"]=None):
        """
        Initializes an instance of a class.
        Args:
            content: Union[str,dict,Message]: Message content. Default: None
        """
        if isinstance(content, dict):
            self.from_dict(content)
            return
        elif isinstance(content, str):
            self.from_str(content)
            return
        elif isinstance(content, Message):
            self.from_message(content)
        else:
            raise TypeError("Parameter 'content' must be either str or dict.")
    
    def from_dict(self, dct: dict) -> None:
        for attr in self._attributes:
            self.__setattr__(attr, dct.get(attr))
        self.set_datetime()
        return
    
    def from_str(self, content: str) -> None:
        for attr in self._attributes:
            self.__setattr__(attr, None)
        self.content = content
        self.role = "user"
        self.set_datetime()
        return
    
    def from_message(self, message: "Message"=None) -> None:
        self.from_dict(message.to_dict)
        return
    
    def load(self, path: str) -> None:
        with open(path, "r", encoding="utf-8") as json_file:
            json_message = json.load(json_file)
        self.from_dict(json_message)
        return
    
    def set_datetime(self) -> None:
        if self.date is None or self.time is None:
            now = datetime.now()
            self.date = now.date().strftime("%Y-%m-%d")
            self.time = now.time().strftime("%H:%M")
        return
    
    def to_openai(self) -> dict:
        return {key: self[key] for key in self._bot_data}
    
    def to_dict(self) -> dict:
        return {key: self[key] for key in self._attributes}
    
    def to_front(self) -> dict:
        return {key: self[key] for key in self._front_keys}
    
    def get(self, key: str) -> Any:
        return self.to_dict().get(key)
            
    def __delitem__(self, item: str) -> None:
        raise KeyError("Access denied.")
            
    def __setitem__(self, item: str, value: Any) -> None:
        self.__setattr__(item, value)
        
    def __getitem__(self, item: str) -> Any:
        return self.__getattribute__(item)
        
    def __delattr__(self, attr: str) -> None:
        raise AttributeError("Access denied.")
    
    def __len__(self) -> int:
        return len(self.content)
    
    def __add__(self, other: Any) -> str:
        return self.content + str(other)
    
    def __radd__(self, other: Any) -> str:
        return self.__add__(other)
    
    def __iadd__(self, other: Any):
        self.content += str(other)
        return self
    
    def __eq__(self, other: Any) -> bool:
        return self.content == str(other)
    
    def __neq__(self, other: Any) -> bool:
        return not self.__eq__(other)
            
    def __repr__(self):
        return str(self.__dict__)
    
    def __str__(self):
        return self.content


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Any
from collections import defaultdict


class RDict(defaultdict):
    """
    RAI wrapper class for defaultdict, which implements attribute addressing
    in addition to item addressing.
    Example:
        dct = RDict(str)
        dct["a"] = 123
        print(dct.a)
        >>> 123
    Note:
        Do NOT use basic method names as the keys, because it will not work properly if
        you address to them via attribute.
        Example:
            dct = RDict(str)
            dct["keys"] = 1234
            print(dct.keys)
            >>> <function RDict.keys> # There exists a default dict method, called 'keys'
        To obtain basic attributes' and methods' names use this:
            print(dir(defaultdict()))
        Or:
            print(defaultdict().__dict__)
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._internal_attribute_names_set = set(dir(self))
        return
    
    def __setattr__(self, attr: str, value: Any) -> None:
        """
        Override a __setattr__ method, so that it sets an item to the dictionary.
        """
        if attr == "_internal_attribute_names_set":
            super().__setattr__(attr, value)
            return
        if attr in self._internal_attribute_names_set:
            raise AttributeError(f"Can't assign value to attribute {attr}. Access denied.")
        self[attr] = value
        return
    
    def __getattr__(self, attr: str) -> Any:
        """
        This method is called whenever you request an attribute that does not exist.
        It returns an item from the dict with the same key, if such a key exists in a dict.
        Otherwise it raises AttributeError.
        """
        if attr in self.keys():
            return self.get(attr)
        else:
            raise AttributeError("Attribute does not exist.")


# Created by: Ausar686
# https://github.com/Ausar686

from .base_actor import BaseActor
from .qagpt import QAGPT
from ..api_pmc_requests import PMCRequester

class KnowledgeBaseSearcher(BaseActor):
    
    _defaults = {
        "ip": "83.220.174.161",
        "port": "9202",
        "prompt": f"Напиши названия 10 статей PubMed Central, в которых подтверждаются следующие слова:\n",
        "url_prefix": "https://pubmed.ncbi.nlm.nih.gov/",
        "descr": "Подробнее об этом можно прочитать в статьях:\n"
    }
    
    def __init__(
        self, 
        model: str="gpt-3.5-turbo",
        *, 
        ip: str=None,
        port: str=None,
        prompt: str=None,
        url_prefix: str=None):
        """
        Initializes an instance of the class
        Args:
            model [str]: GPT model. Available: "gpt-3.5-turbo", "gpt-4". Default: "gpt-3.5-turbo".
        Kwargs:
            ip [str]: Knowledge base IP for PMCR. If None is passed, uses Heuristics PMC IP. Default: None.
            port [str]: Knowledge base port for PMCR. If None is passed, uses Heuristics PMC port. Default: None.
            prompt [str]: Prompt to generate relevant articles' titles. If None is passed, uses default prompt. Default: None.
            url_prefix [str]: URL prefix for URL-generation from PMID (PMC-specific). If Non is passed, uses default prefix. Default: None.
        """
        # Initialize required parameters
        super().__init__(model)
        if ip is None:
            self.ip = self._defaults["ip"]
        else:
            self.ip = ip
        if port is None:
            self.port = self._defaults["port"]
        else:
            self.port = port
        if prompt is None:
            self.prompt = self._defaults["prompt"]
        else:
            self.prompt = prompt
        # Initialize PMC-specific parameters
        self.url_prefix = self._defaults["url_prefix"]
        self.descr = self._defaults["descr"]
        # Initialize GPT instance
        self.gpt = QAGPT(self.model)
        # Initialize PMCR instance
        self.pmc = PMCRequester(ip_address=self.ip, port=self.port)
        return
    
    def wrap(self, text: str) -> str:
        """
        Wrap an input string to further address it to OpenAI API.
        """
        return f"{self.prompt}{text}"
    
    def generate(self, text: str=None) -> list:
        """
        Generates names of the relevant articles with QAGPT.
        """
        if text is None or text == "":
            return []
        request = self.wrap(text)
        articles = self.gpt.get_list(request)
        return articles
    
    def filter(self, articles: list=None) -> list:
        """
        Filters generated names of the articles with PMCRequester.
        Returns a list of tuples: (article_title, article_url)
        """
        if articles is None or articles == []:
            return []
        filtered_articles = []
        for article in articles:
            title, pmid = self.pmc.find_title(article)
            if title is not None:
                url = self.url_prefix + pmid
                filtered_articles.append((title, url))
        return filtered_articles
    
    def to_json(self, text: str=None, filtered_articles: list=None) -> dict:
        """
        Converts search results into JSON format.
        """
        if text is None or text == "":
            return {}
        if filtered_articles is None or filtered_articles == []:
            return {"text": text, "descr": self.descr, "article": []}
        return {
            "text": text, 
            "descr": self.descr, 
            "article": [{"title": title, "url": url} for (title, url) in filtered_articles]
        }
    
    def run(self, text: str=None) -> dict:
        """
        Executes full pipeline: 
            1. Generates relevant article titles with QAGPT
            2. Filters generated titles with PMCR
            3. Converts the results into JSON format (Python dict)
        """
        articles = self.generate(text)
        filtered_articles = self.filter(articles)
        json_data = self.to_json(text, filtered_articles)
        return json_data
    
    @staticmethod
    def to_str(json_data: dict) -> str:
        """
        Converts the result of 'run' method into string format for further display to user.
        """
        text = json_data.get("text", "").strip() + "\n"
        descr = json_data.get("descr", "").strip() + "\n\n"
        articles = json_data.get("article", [])
        article_str = ""
        for i, article in enumerate(articles, start=1):
            article_str += f"{i}.{article.get('title', '')}\n{article.get('url', '')}\n\n"
        res = f"{text}{descr}{article_str}"
        return res


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union
from collections import deque

from .base_actor import BaseActor
from .token_counter import TokenCounter
from .qagpt import QAGPT


class TextSummarizer(BaseActor):
    
    # These constants can be edited
    # if they prove to be non-sufficient for output storage.
    _limits = {
        "gpt-3.5-turbo-16k": 15500,
        "gpt-4-32k": 31500
    }
    
    _upgrades = {
        "gpt-3.5-turbo": "gpt-3.5-turbo-16k",
        "gpt-4": "gpt-4-32k"
    }
    
    _max_words = 200
    
    def __init__(self, model: str="gpt-3.5-turbo", n_words: int=50):
        super().__init__(model)
        if n_words > self._max_words:
            raise ValueError("Too many words for summary. Maximum 200 words allowed.")
        self.n_words = n_words
        self.token_counter = TokenCounter(self.model)
        self.gpt = QAGPT(model=self.model)
        return

    def set_model(self, model: str) -> None:
        super().set_model(model)
        self.upgrade_model()
        return
    
    def upgrade_model(self) -> None:
        # Upgrades model to the one with higher token cap
        # If it's not possible, does nothing
        try:
            self.model = self._upgrades[self.model]
            return
        except KeyError:
            pass
    
    @property
    def prompt(self):
        prompt = f"""
            <INSTRUCTION>
            Напиши краткое содержание текста ниже.
            В кратком содержании должно быть максимум {self.n_words} слов.
            Ответ выпиши на русском языке.
            """
        return prompt
    
    @staticmethod
    def dict2str(dct: dict) -> str:
        # Converts message in a dict form to message in a dialog (string) form
        return f"[{dct.get('role')}]: {dct.get('content')}"
        
    def run(self, obj: Union[str, list, dict]) -> str:
        # Raw text for summariztaion
        if isinstance(obj, str):
            return self.summarize_str(obj)
        # Message dict for summarization
        elif isinstance(obj, dict):
            return self.summarize_dict(obj)
        # List of messages for summarization
        elif isinstance(obj, list):
            return self.summarize_list(obj)
        elif isinstance(obj, deque):
            # If a deque is given, process it as a list.
            return self.summarize_list(obj)
        else:
            raise TypeError(f"Unsupported type {type(obj)} for argument 'obj'.",
                            f"Only str, dict and list are supported.")

    def wrap(self, string: str) -> str:
        return self.prompt + f"""
            <TEXT>
            {string}
            """
        
    def summarize_str(self, string: str) -> str:
        request = self.wrap(string)
        n_tokens = self.token_counter.run(request)
        # If text is super long, split it in 2 parts
        if n_tokens > self._limits[self.model]:
            lines = request.split("\n")
            n_lines = len(lines)
            batch = "\n".join(lines[:n_lines//2])
            n_tokens_batch = self.token_counter.run(batch)
            if n_tokens_batch > self._limits[self.model]:
                # Probably, it's a very long single message.
                # Summarize it in a bruteforce manner.
                return self.brute_summarize(string)
            summary = self.summarize_str(batch)
            new_string = summary + "\n" + "\n".join(lines[n_lines//2:])
            return self.summarize_str(new_string)
        answer = self.gpt.get_str(request)
        return answer

    def brute_summarize(self, string: str) -> str:
        # Split large string with no newlines into two parts
        # Summarize first one and add the result to second one
        # Then summarize the mixed string
        size = len(string)
        batch = string[:size//2]
        summary = self.summarize_str(batch)
        new_string = summary + string[size//2:]
        return self.summarize_str(new_string)
    
    def summarize_dict(self, dct: dict) -> str:
        # Convert message in a dict form to message in a dialog (string) form.
        # And then call 'summarize_str'
        string = self.dict2str(dct)
        return self.summarize_str(string)
    
    def summarize_list(self, lst: list) -> str:
        # Convert all messages from the list (dicts) to strings
        # Then join the strings via newline
        # And then call 'summarize_str'
        strings = [self.dict2str(dct) for dct in lst]
        string = "\n".join(strings)
        return self.summarize_str(string)


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Any, Union, Tuple, Iterable
from datetime import datetime

import pandas as pd
from pandas.core.frame import DataFrame
from pandas.core.series import Series


class BaseStorage:
    _base_columns = ["date", "time", "datetime"]
    _base_indexed_columns = ["date"]
    _base_multivalue_columns = []
    _columns = []
    _indexed_columns = []
    _multivalue_columns = []
    
    def __init__(self, df: Union[DataFrame,str]=None):
        self._index_col = "index"
        self._columns.extend(self._base_columns)
        self._indexed_columns.extend(self._base_indexed_columns)
        self._multivalue_columns.extend(self._base_multivalue_columns)
        if df is None:
            self.data = pd.DataFrame(columns=self._columns)
            return
        elif isinstance(df, str):
            self.from_csv(df)
        else:
            self.data = df.copy()
        extra_columns = [column for column in self.data.columns if column not in self._columns]
        self.data.drop(extra_columns, axis=1, inplace=True)
        for column in self._columns:
            if column == "datetime":
                self.data["datetime"] = pd.to_datetime(self.data["date"] + ' ' + self.data["time"])
            if column not in self.data.columns:
                self.data[column] = None
        return
    
    def create(self, dct: dict) -> dict:
        note = {key: self.iter_to_str(value) if key in self._multivalue_columns
                else value for key, value in dct.items()}
        now = datetime.now()
        note["date"] = now.date().strftime("%Y-%m-%d")
        note["time"] = now.time().strftime("%H:%M")
        note["datetime"] = note["date"] + ' ' + note["time"]
        return note
    
    def add(self, note: dict) -> None:
        new_row = pd.Series(note)
        try:
            index = self.data.index[-1] + 1
        except IndexError:
            index = 0
        self.data.loc[index] = new_row
        return
    
    def append(self, dct: dict) -> None:
        # A convinient way to add a new note from the dict to the storage.
        # This method consequently performs 'create' and 'add' methods.
        note = self.create(dct)
        self.add(note)
        return
    
    def delete(self, index: int) -> None:
        try:
            self.data.drop(index, inplace=True)
        except KeyError:
            print(f"[INFO]: No row with index {index} found. Ignored.")
        return
    
    def edit(self, index: int=None, note: Union[dict, Series]=None) -> None:
        if index is None and note is None:
            return
        # If index is None, we are trying to edit a specific row, metioned in note
        if index is None: 
            index = note.name
        if index not in self.data.index:
            print(f"[INFO]: No row with index {index} found. Ignored.")
            return
        for key in note:
            if key not in self._columns:
                continue
            value = note[key]
            if key in self._multivalue_columns and self.is_nonstr_iterable(value):
                value = self.iter_to_str(value)
            self.data.loc[index, key] = value
        return
    
    def _base_get(self, 
                  *, 
                  by: str,
                  value: Any=None,
                  search_type: str="union",
                  date_search: str="precise") -> Tuple[bool, Union[DataFrame, Series]]:
        # Performs basic search, which is similar to all subclasses.
        # 'value' is not specified. Return all data 
        if value is None:
            return True, self.data
        # Search by index via 'loc' method
        elif by == "index":
            return True, self.data.loc[value]
        elif by == "date":
            if date_search == "precise":
                return True, self._eq_search(by, value)
            elif date_search == "between":
                date_start = value[0]
                date_end = value[1]
                return True, self._date_search(date_start, date_end)
            else:
                raise ValueError("Parameter 'date_search' must be either 'precise' or 'between'.")
        # 'by' is either a string or non-iterable object
        # and 'value' is a non-string iterable object
        elif not self.is_nonstr_iterable(by) and self.is_nonstr_iterable(value):
            if search_type not in ["union", "intersection"]:
                raise ValueError("Parameter 'search_type' must be either 'union' or 'intersection'.")
            res_idx = None
            for i, value_ in enumerate(value):
                # Iterate over values in 'value'
                # And apply index union/intersection among results
                # The type of operation depends on 'search_type' parameter
                current_res = self.get(by=by, value=value_)
                if i == 0:
                    res_idx = current_res.index
                else:
                    res_idx = res_idx.__getattribute__(search_type)(current_res.index)
            # Return rows with indexes in 'res_idx'
            return True, self.data.loc[res_idx]
        elif self.is_nonstr_iterable(by):
            # Group arguments into pairs column-value
            pairs = zip(by, value)
            res_idx = None
            for i, pair in enumerate(pairs):
                # Iterate over pairs in zip
                # And apply index intersection among results
                by_, value_ = pair
                current_res = self.get(by=by_, value=value_)
                if i == 0:
                    res_idx = current_res.index
                else:
                    res_idx = res_idx.intersection(current_res.index)
            # Return rows with indexes in 'res_idx'
            return True, self.data.loc[res_idx]
#             raise NotImplementedError("Multiple search will be implemented in upcoming updates.")
        elif by not in self._indexed_columns:
            raise ValueError("Invalid search argument.")
        return False, None
            
    def _eq_search(self, column: str, value: Any) -> DataFrame:
        # Columnwise equality-based search.
        return self.data[self.data[column]==value]
    
    def _substr_search(self, column: str, value: Any) -> DataFrame:
        # Return all rows, where the following column cell
        return self.data[self.data[column].str.contains(value)]
    
    def _date_search(self, date_start: str, date_end: str):
        # Return all rows, where date is in between 'date_start' and 'date_end'
        return self.data[(date_start <= self.data.date) & (self.data.date <= date_end)]
    
    def get(self, 
            *,
            by: str,
            value: Any=None,
            search_type: str="union",
            date_search: str="precise") -> DataFrame:
        # Define a 'get' method for subclasses.
        # Implementation depends on a specific subclass.
        # It is strongly recommended to use '_base_get' method
        # in the beginning of 'get' method in any subclass
        pass
    
    @staticmethod
    def is_iterable(obj: Any) -> bool:
        try:
            iter(obj)
            return True
        except TypeError:
            return False
        
    @classmethod
    def is_nonstr_iterable(cls, obj: Any) -> bool:
        return cls.is_iterable(obj) and not isinstance(obj, str)
    
    @classmethod
    def iter_to_str(cls, lst: Iterable[str]) -> str:
        if cls.is_nonstr_iterable(lst):
            return '|'.join(lst)
        return lst

    def from_csv(self, path: str) -> None:
        self.data = pd.read_csv(path, index_col=self._index_col)
        return
    
    def to_csv(self, path: str) -> None:
        self.data.drop_duplicates(self._columns, keep="last").to_csv(path, index_label=self._index_col)
        return
    
    def load(self, path: str) -> None:
        # An alias for 'from_csv' method for standartized interface
        self.from_csv(path)
        return
    
    def to_disk(self, path: str) -> None:
        # An alias for 'to_csv' method for standartized interface
        self.to_csv(path)
        return
        
    def __repr__(self):
        return repr(self.data)
    
    def __str__(self):
        return str(self.data)


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union
from datetime import datetime
import os
import hashlib
import copy
import json

from .message import Message


class Chat:
    
    _image_dir = "bot_images"
    _data_dir = "chats"
    _default_name = "chat"
    _chunk_size = 100
    _suffix_len = 7
    _attributes = ["username", "bot_name", "messages", "recent_messages", "image"]
    _names = ["username", "bot_name"]
    _db_columns = ["chat_id", "role", "content", "date", "time"]
    _db_table = "messages"
    
    def __init__(self, load_data: Union[str,dict]=None):
        if isinstance(load_data, str):
            self.from_path(load_data)
            return
        elif isinstance(load_data, dict):
            self.from_dict(load_data)
            return
        else:
            raise TypeError("Parameter 'load_data' must be either dict or str.")
    
    def from_dict(self, dct: dict) -> None:
        for attr in self._attributes:
            if attr == "messages":
                self.__setattr__(attr, copy.deepcopy(dct.get(attr)))
                if self.messages is None:
                    self.messages = []
            elif attr == "recent_messages":
                self.recent_messages = []
            else:
                self.__setattr__(attr, dct.get(attr))
        for name in self._names:
            if self.__getattribute__(name) is None:
                raise ValueError("Username and bot name must be not None.")
        self.chat_id = self.make_id()
        for message in self.messages:
            self.set_message_data(message, setup_datetime=False)
        self.image = self.get_image_path()
        return
    
    def from_path(self, load_dir: str=None) -> None:
        if load_dir is None:
            raise ValueError("Directory must be not None.")
        if not os.path.exists(load_dir):
            raise ValueError("Directory does not exist.")
        load_dir = os.sep.join(load_dir.split('/'))
        self.bot_name = load_dir.split(os.sep)[-1]
        self.username = load_dir.split(os.sep)[-2]
        self.chat_id = self.make_id()
        self.messages = []
        self.recent_messages = []
        for filename in os.listdir(load_dir):
            path = os.path.join(load_dir, filename)
            with open(path, "r", encoding="utf-8") as json_file:
                json_messages = json.load(json_file)
            messages = [Message(json_message) for json_message in json_messages]
            self.messages.extend(messages)
        for message in self.messages:
            self.set_message_data(message, setup_datetime=False)
        self.image = self.get_image_path()
        return
    
    def get_image_path(self) -> str:
        return os.path.join(self._image_dir, f"{self.bot_name}.jpg")
    
    def make_id(self) -> str:
        # Returns an id of the chat by hashing data
        return hashlib.sha256((f"{self.username}###{self.bot_name}").encode()).hexdigest()
    
    def set_message_data(self, message: Message=None, setup_datetime: bool=True) -> None:
        if message is None:
            return
        if message.role == "assistant":
            message.username = self.bot_name
        elif message.role == "user":
            message.username = self.username
        else:
            message.username = "ROOT"
        message.chat_id = self.chat_id
        if setup_datetime:
            now = datetime.now()
            message.date = now.date().strftime("%Y-%m-%d")
            message.time = now.time().strftime("%H:%M")
        return
        
    def append(self, message: Union[Message,dict]=None) -> None:
        if message is None:
            return
        if isinstance(message, dict):
            message = Message(message)
        self.set_message_data(message)
        self.messages.append(message)
        self.recent_messages.append(message)
        return
    
    @property
    def last(self) -> Message:
        if self.messages:
            return self.messages[-1]
        return None
    
    def to_front(self) -> dict:
        if self.last is not None:
            front_dict = self.last.to_front()
            front_dict["image"] = self.image
            return {self.bot_name: front_dict}
        else:
            return {self.bot_name: {key: None for key in Message._front_keys}}
        
    def to_database(self):
        # TODO: Write actual code for this
        # TODO: Make this method async
        # Sends messages to database
        self.recent_messages.clear()
        return
        
    def __len__(self) -> int:
        return len(self.messages)
    
    def __iter__(self):
        return iter(self.messages)
    
    def __getitem__(self, item: int) -> Message:
        return self.messages[item]
        
    def load(self, load_dir: str=None) -> None:
        # An alias for 'from_path' method for standartized interface
        self.from_path(load_dir)
        return
    
    def to_disk(self) -> None:
        output_dir = self._data_dir
        # if not os.path.exists(output_dir):
            # raise ValueError("Directory does not exist.")
        output_user_dir = os.path.join(output_dir, self.username)
        # if not os.path.exists(output_user_dir):
        #     os.mkdir(output_user_dir)
        output_bot_dir = os.path.join(output_user_dir, self.bot_name)
        if not os.path.exists(output_bot_dir):
            os.makedirs(output_bot_dir)
        splitted_messages = []
        for i in range(0, len(self), self._chunk_size):
            chunk = [message.to_dict() for message in self.messages[i: i+self._chunk_size]]
            splitted_messages.append(chunk)
        for i, chunk in enumerate(splitted_messages):
            suffix = str(i).zfill(self._suffix_len)
            filename = f"{self._default_name}{suffix}.json"
            path = os.path.join(output_bot_dir, filename)
            with open(path, "w", encoding="utf-8") as json_file:
                json.dump(chunk, json_file)
        return


# Created by: Ausar686
# https://github.com/Ausar686

import json
import re

from ..actors.qagpt import QAGPT


class PollGenerator:
    
    def __init__(self, sphere: str, example_path: str=None, log: bool=True) -> None:
        if example is not None:
            with open(example_path, "r", encoding="utf-8") as file:
                self.example = file.read()
        else:
            self.example = ''
        self.sphere = sphere
        self.gpt = QAGPT()
        self.log = log
        self.topics = []
        self.polls = {}
        self.attempts = []
        return
        
    def print(self, *args, **kwargs)-> None:
        if self.log:
            print(*args, **kwargs)
        return
        
    def get_topics(self) -> None:
        request = f"Напиши список основных тем опросников в сфере '{self.sphere}'"
        self.topics = self.gpt.get_list(request)
        nl = '\n'
        self.print(f"[INFO]: Successfully obtained a list of topics for sphere '{self.sphere}'")
        self.print(f"Here they are:\n{nl.join(self.topics)}")
        return
        
    def run(self, output: bool=False) -> None:
        self.attempts.clear()
        self.get_topics()
        for topic in self.topics:
            while True:
                request = f"""
                    Напиши опрос на тему {topic}.
                    Опрос должен состоять из 7 вопросов.
                    Вкаждом вопросе должно быть 3-4 варианта ответа.
                    Опрос должен быть представлен в формате JSON.
                    Каждый ключ в JSON-файле должен быть внутри двойных кавычек.
                    Пример формата указан ниже:
                    {self.example}
                """
                answer = self.gpt.get_str(request)
                self.attempts.append(answer)
                new_answer = re.sub(r"\s+", " ", answer)
                try:
                    poll = json.loads(new_answer)
                    self.polls[topic] = poll
                    self.print(f"[INFO]: Successfully created poll on topic: '{topic}'")
                    if output:
                        self.to_json(topic, f"{topic}.json")
                    break
                except Exception as e:
                    self.print(f"[ERROR]: An exception has occured during generation poll on topic '{topic}'.")
                    self.print(f"'{e}'. Retrying...")
        self.print(f"[INFO]: Finished generating polls for sphere: '{self.sphere}'.")
        return
            
    def to_json(self, topic: str, path: str) -> None:
        if topic not in self.polls:
            return
        with open(path, "w") as json_file:
            json.dump(self.polls[topic], json_file)
        self.print(f"[INFO]: Successfully written poll '{topic}' into a file: '{path}'")
        return


# Created by: Ausar686
# https://github.com/Ausar686

class BaseActor:
	"""
	Base class for all RAI actors, which can be used in ChatBot
	"""

	def __init__(self, model: str="gpt-3.5-turbo"):
		"""
		Initialize an actor.
		"""
		self.set_model(model)
		return

	def set_model(self, model: str) -> None:
		"""
		Set actor's model.
		"""
		self.model = model
		return

	def run(self):
		"""
		Override this method for every subclass in order to use it properly as an actor in ChatBot.
		"""
		pass


# Created by: Ausar686
# https://github.com/Ausar686

def method_logger(func):
    """
    Special decorator for logging RAI ChatBot methods.
    """
    
    def start_log(func_name, args, kwargs):
        strings = [
            f"[INFO]: EXECUTING: {func.__name__}",
            f"[INFO]: Args: {args}",
            f"[INFO]: Kwargs: {kwargs}"
        ]
        print("\n".join(strings))
        return
    
    def end_log(func_name, args, kwargs):
        strings = [
            f"[INFO]: FINISHED: {func.__name__}",
            f"[INFO]: Args: {args}",
            f"[INFO]: Kwargs: {kwargs}"
        ]
        print("\n".join(strings))
        return
        
    def safe_log(logger, obj, func_name, args, kwargs):
        try:
            if obj.log:
                logger(func_name, args, kwargs)
        except Exception:
            pass
        return
    
    def wrapper(*args, **kwargs):
        obj = args[0]
        safe_log(start_log, obj, func.__name__, args, kwargs)
        res = func(*args, **kwargs)
        safe_log(end_log, obj, func.__name__, args, kwargs)
        return res
    
    return wrapper


def retry(n_retries: int=5):
    """
    Retrying decorator function.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            n = n_retries
            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    n -= 1
                    if n <= 0:
                        raise
        return wrapper
    return decorator


from setuptools import setup, find_packages

setup(name="redemption-ai",
      version="0.0.2",
      description="AI-powered (OpenAI API) tools for chat-like interactions and personal AI assistance.",
      packages=find_packages(include=["RAI", "RAI.*"]),
      author="Ausar686",
      author_email='glebyushkov@mail.ru',
      install_requires=[
        "numpy==1.25.2",
        "openai==0.28.0",
        "orjson==3.9.7",
        "pandas==2.1.0",
        "pydantic==2.3.0",
        "requests==2.31.0",
        "tiktoken==0.5.0"
      ])


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union
from collections import deque

import tiktoken

from .base_actor import BaseActor
from ..chat import Message


class TokenCounter(BaseActor):
    _tokens_per_message = 3
    _models = [
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-16k",
        "gpt-4",
        "gpt-4-32k",
    ]
    
    def __init__(self, model: str="gpt-3.5-turbo"):
        super().__init__(model)
        self.encodings = {key: tiktoken.encoding_for_model(key) for key in self._models}
        return

    @property
    def encoding(self):
        return self.encodings[self.model]
    
    def run(self, obj: Union[str, dict, list, None]) -> int:
        if obj is None:
            return 0
        if isinstance(obj, str):
            return self.count_from_str(obj)
        elif isinstance(obj, dict):
            return self.count_from_dict(obj)
        elif isinstance(obj, list):
            return self.count_from_list(obj)
        elif isinstance(obj, deque):
            # Counting for list and deque are the same
            return self.count_from_list(obj)
        elif isinstance(obj, Message):
            # Counting for Message is equal to counting from dict
            return self.count_from_dict(obj.to_openai())
        else:
            raise TypeError(f"Parameter 'obj' must be str, dict, list, deque, Message or None, not {type(obj)}.")
            
    def count_from_str(self, string: str) -> int:
        return len(self.encoding.encode(string))
    
    def count_from_dict(self, dct: dict) -> int:
        # Here we assume, that a message is provided in OpenAI dict form:
        # {"role": role, "content": content}
        # So we simply call 'count_from_str' on content
        return self.count_from_str(dct.get("content"))
    
    def count_from_list(self, lst: list) -> int:
        # Here we assume that a list of messages in OpenAI dict form is given:
        # [{"role": role, "content": content}, ...]
        # So we simply iterate over the list and call 'count_from_dict'
        num_tokens = 0
        for message in lst:
            num_tokens += self._tokens_per_message
            num_tokens += self.count_from_dict(message)
        num_tokens += 3 # every reply is primed with <|start|>assistant<|message|>
        return num_tokens